@@ -3,61 +3,27 @@
 # This source code is licensed under the MIT license found in the
 # LICENSE file in the root directory of this source tree.
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
 from contextlib import contextmanager
 
-from typing import List  # noqa
 import sys
+import asyncio.log
 import unittest
+import contextlib
 import re
 import types
+import asyncio
+import inspect
+import warnings
 
 import testslide.mock_callable
 import testslide.mock_constructor
+import testslide.matchers
+import testslide.patch_attribute
 from testslide.strict_mock import StrictMock  # noqa
 
-if sys.version_info.major >= 3 and sys.version_info.minor >= 5:
-    from contextlib import redirect_stdout, redirect_stderr
-else:
 
-    @contextmanager
-    def redirect_stdout(target):
-        original = sys.stdout
-        sys.stdout = target
-        try:
-            yield
-        finally:
-            sys.stdout = original
-
-    @contextmanager
-    def redirect_stderr(target):
-        original = sys.stderr
-        sys.stderr = target
-        try:
-            yield
-        finally:
-            sys.stderr = original
-
-
-@contextmanager
-def _add_traceback_context_manager():
-    """
-    Add __traceback__ to exceptions raised within the block, to give Python
-    2 compatibility.
-    """
-    if sys.version_info[0] == 2:
-        try:
-            yield
-        except BaseException as e:
-            _exc_type, _exc_value, exc_traceback = sys.exc_info()
-            e.__dict__["__traceback__"] = exc_traceback
-            raise e
-    else:
-        yield
+if sys.version_info < (3, 6):
+    raise RuntimeError("Python >=3.6 required.")
 
 
 def _importer(target):
@@ -84,32 +50,63 @@ class _ContextData(object):
     example execution.
     """
 
-    def __init__(self, context):
-        self.context = context
-        self.after_functions = []
-        # The use of methodName parameter is required as a placeholder for
-        # Python 2 compatibility only.
-        method_name = "assertEqual"
-        self._test_case = unittest.TestCase(methodName=method_name)
+    def _init_sub_example(self):
         self._sub_examples_agg_ex = AggregatedExceptions()
 
-        def assert_sub_examples(self):
+        def real_assert_sub_examples(self):
             if self._sub_examples_agg_ex.exceptions:
                 self._sub_examples_agg_ex.raise_correct_exception()
 
+        if self._example.is_async:
+
+            async def assert_sub_examples(self):
+                real_assert_sub_examples(self)
+
+        else:
+
+            def assert_sub_examples(self):
+                real_assert_sub_examples(self)
+
         self.after(assert_sub_examples)
 
+    def _init_mocks(self):
+        self.mock_callable = testslide.mock_callable.mock_callable
+        self.mock_async_callable = testslide.mock_callable.mock_async_callable
+        self.mock_constructor = testslide.mock_constructor.mock_constructor
+        self.patch_attribute = testslide.patch_attribute.patch_attribute
+        self._mock_callable_after_functions = []
+
+        def register_assertion(assertion):
+            if self._example.is_async:
+
+                async def f(_):
+                    assertion()
+
+            else:
+                f = lambda _: assertion()
+            self._mock_callable_after_functions.append(f)
+
+        testslide.mock_callable.register_assertion = register_assertion
+
+    def __init__(self, example):
+        self._example = example
+        self._context = example.context
+        self._after_functions = []
+        self._test_case = unittest.TestCase()
+        self._init_sub_example()
+        self._init_mocks()
+
     @staticmethod
     def _not_callable(self):
         raise BaseException("This function should not be called outside test code.")
 
     @property
     def _all_methods(self):
-        return self.context.all_context_data_methods
+        return self._context.all_context_data_methods
 
     @property
     def _all_attributes(self):
-        return self.context.all_context_data_memoizable_attributes
+        return self._context.all_context_data_memoizable_attributes
 
     def __getattr__(self, name):
         if name in self._all_methods.keys():
@@ -120,7 +117,12 @@ class _ContextData(object):
             self.__dict__[name] = static
 
         if name in self._all_attributes.keys():
-            self.__dict__[name] = self._all_attributes[name](self)
+            attribute_code = self._all_attributes[name]
+            if self._example.is_async and inspect.iscoroutinefunction(attribute_code):
+                raise ValueError(
+                    f"Function can not be a coroutine function: {repr(attribute_code)}"
+                )
+            self.__dict__[name] = attribute_code(self)
 
         try:
             return self.__dict__[name]
@@ -129,7 +131,7 @@ class _ContextData(object):
             if re.match("^assert", name) and hasattr(self._test_case, name):
                 return getattr(self._test_case, name)
             raise AttributeError(
-                "Context '{}' has no attribute '{}'".format(self.context, name)
+                "Context '{}' has no attribute '{}'".format(self._context, name)
             )
 
     def after(self, after_code):
@@ -137,7 +139,7 @@ class _ContextData(object):
         Use this to decorate a function to be registered to be executed after
         the example code.
         """
-        self.after_functions.append(after_code)
+        self._after_functions.append(after_code)
         return self._not_callable
 
     @contextmanager
@@ -169,13 +171,14 @@ class AggregatedExceptions(Exception):
     @contextmanager
     def catch(self):
         try:
-            with _add_traceback_context_manager():
-                yield
+            yield
         except BaseException as exception:
             self.append_exception(exception)
 
     def __str__(self):
-        return "{} failures.".format(len(self.exceptions))
+        return "{} failures.\n".format(len(self.exceptions)) + "\n".join(
+            f"{type(e)}: {str(e)}" for e in self.exceptions
+        )
 
     def raise_correct_exception(self):
         if not self.exceptions:
@@ -183,8 +186,14 @@ class AggregatedExceptions(Exception):
         ex_types = {type(ex) for ex in self.exceptions}
         if Skip in ex_types or unittest.SkipTest in ex_types:
             raise Skip()
+        elif len(self.exceptions) == 1:
+            raise self.exceptions[0]
         else:
             raise self
+            if len(self.exceptions) == 1:
+                raise self.exceptions[0]
+            else:
+                raise self
 
 
 class Skip(Exception):
@@ -201,6 +210,214 @@ class UnexpectedSuccess(Exception):
     """
 
 
+class SlowCallback(Exception):
+    """
+    Raised by TestSlide when an asyncio slow callback warning is detected
+    """
+
+
+class _ExampleRunner:
+    def __init__(self, example):
+        self.example = example
+
+    @staticmethod
+    async def _fail_if_not_coroutine_function(func, *args, **kwargs):
+        if not inspect.iscoroutinefunction(func):
+            raise ValueError(f"Function must be a coroutine function: {repr(func)}")
+        return await func(*args, **kwargs)
+
+    async def _real_async_run_all_hooks_and_example(
+        self, context_data, around_functions=None
+    ):
+        """
+        ***********************************************************************
+        ***********************************************************************
+                                    WARNING
+        ***********************************************************************
+        ***********************************************************************
+
+        This function **MUST** be keep the exact same execution flow of
+        _sync_run_all_hooks_and_example()!!!
+        """
+        if around_functions is None:
+            around_functions = list(reversed(self.example.context.all_around_functions))
+
+        if not around_functions:
+            aggregated_exceptions = AggregatedExceptions()
+            with aggregated_exceptions.catch():
+                for before_code in self.example.context.all_before_functions:
+                    await self._fail_if_not_coroutine_function(
+                        before_code, context_data
+                    )
+                await self._fail_if_not_coroutine_function(
+                    self.example.code, context_data
+                )
+            after_functions = []
+            after_functions.extend(context_data._mock_callable_after_functions)
+            after_functions.extend(self.example.context.all_after_functions)
+            after_functions.extend(context_data._after_functions)
+            for after_code in reversed(after_functions):
+                with aggregated_exceptions.catch():
+                    await self._fail_if_not_coroutine_function(after_code, context_data)
+            aggregated_exceptions.raise_correct_exception()
+            return
+        around_code = around_functions.pop()
+
+        wrapped_called = []
+
+        async def async_wrapped():
+            wrapped_called.append(True)
+            await self._real_async_run_all_hooks_and_example(
+                context_data, around_functions
+            )
+
+        await self._fail_if_not_coroutine_function(
+            around_code, context_data, async_wrapped
+        )
+
+        if not wrapped_called:
+            raise RuntimeError(
+                "Around hook "
+                + repr(around_code.__name__)
+                + " did not execute example code!"
+            )
+
+    @contextlib.contextmanager
+    def _raise_if_asyncio_warnings(self, context_data):
+        if sys.version_info < (3, 7):
+            yield
+            return
+        original_showwarning = warnings.showwarning
+        caught_failures = []
+
+        def showwarning(message, category, filename, lineno, file=None, line=None):
+            failure_warning_messages = {
+                RuntimeWarning: "^coroutine '.+' was never awaited"
+            }
+            warning_class = type(message)
+            pattern = failure_warning_messages.get(warning_class, None)
+            if pattern and re.compile(pattern).match(str(message)):
+                caught_failures.append(message)
+            else:
+                original_showwarning(message, category, filename, lineno, file, line)
+
+        warnings.showwarning = showwarning
+
+        original_logger_warning = asyncio.log.logger.warning
+
+        def logger_warning(msg, *args, **kwargs):
+            if re.compile("^Executing .+ took .+ seconds$").match(str(msg)):
+                msg = (
+                    f"{msg}\n"
+                    "During the execution of the async test a slow callback "
+                    "that blocked the event loop was detected.\n"
+                    "Tip: you can customize the detection threshold with:\n"
+                    "  asyncio.get_running_loop().slow_callback_duration = seconds"
+                )
+                caught_failures.append(SlowCallback(msg % args))
+            else:
+                original_logger_warning(msg, *args, **kwargs)
+
+        asyncio.log.logger.warning = logger_warning
+
+        aggregated_exceptions = AggregatedExceptions()
+
+        try:
+            with aggregated_exceptions.catch():
+                yield
+        finally:
+            warnings.showwarning = original_showwarning
+            asyncio.log.logger.warning = original_logger_warning
+            for failure in caught_failures:
+                with aggregated_exceptions.catch():
+                    raise failure
+            aggregated_exceptions.raise_correct_exception()
+
+    def _async_run_all_hooks_and_example(self, context_data):
+        coro = self._real_async_run_all_hooks_and_example(context_data)
+        with self._raise_if_asyncio_warnings(context_data):
+            if sys.version_info < (3, 7):
+                loop = asyncio.events.new_event_loop()
+                try:
+                    loop.set_debug(True)
+                    loop.run_until_complete(coro)
+                finally:
+                    try:
+                        loop.run_until_complete(loop.shutdown_asyncgens())
+                    finally:
+                        loop.close()
+            else:
+                asyncio.run(coro, debug=True)
+
+    @staticmethod
+    def _fail_if_coroutine_function(func, *args, **kwargs):
+        if inspect.iscoroutinefunction(func):
+            raise ValueError(f"Function can not be a coroutine function: {repr(func)}")
+        return func(*args, **kwargs)
+
+    def _sync_run_all_hooks_and_example(self, context_data, around_functions=None):
+        """
+        ***********************************************************************
+        ***********************************************************************
+                                    WARNING
+        ***********************************************************************
+        ***********************************************************************
+
+        This function **MUST** be keep the exact same execution flow of
+        _real_async_run_all_hooks_and_example()!!!
+        """
+        if around_functions is None:
+            around_functions = list(reversed(self.example.context.all_around_functions))
+
+        if not around_functions:
+            aggregated_exceptions = AggregatedExceptions()
+            with aggregated_exceptions.catch():
+                for before_code in self.example.context.all_before_functions:
+                    self._fail_if_coroutine_function(before_code, context_data)
+                self._fail_if_coroutine_function(self.example.code, context_data)
+            after_functions = []
+            after_functions.extend(context_data._mock_callable_after_functions)
+            after_functions.extend(self.example.context.all_after_functions)
+            after_functions.extend(context_data._after_functions)
+            for after_code in reversed(after_functions):
+                with aggregated_exceptions.catch():
+                    self._fail_if_coroutine_function(after_code, context_data)
+            aggregated_exceptions.raise_correct_exception()
+            return
+        around_code = around_functions.pop()
+
+        wrapped_called = []
+
+        def wrapped():
+            wrapped_called.append(True)
+            self._sync_run_all_hooks_and_example(context_data, around_functions)
+
+        self._fail_if_coroutine_function(around_code, context_data, wrapped)
+
+        if not wrapped_called:
+            raise RuntimeError(
+                "Around hook "
+                + repr(around_code.__name__)
+                + " did not execute example code!"
+            )
+
+    def run(self):
+        try:
+            if self.example.skip:
+                raise Skip()
+            context_data = _ContextData(self.example)
+            if self.example.is_async:
+                self._async_run_all_hooks_and_example(context_data)
+            else:
+                self._sync_run_all_hooks_and_example(context_data)
+        finally:
+            sys.stdout.flush()
+            sys.stderr.flush()
+            testslide.mock_callable.unpatch_all_callable_mocks()
+            testslide.mock_constructor.unpatch_all_constructor_mocks()
+            testslide.patch_attribute.unpatch_all_mocked_attributes()
+
+
 class Example(object):
     """
     Individual example.
@@ -209,6 +426,7 @@ class Example(object):
     def __init__(self, name, code, context, skip=False, focus=False):
         self.name = name
         self.code = code
+        self.is_async = inspect.iscoroutinefunction(self.code)
         self.context = context
         self.__dict__["skip"] = skip
         self.__dict__["focus"] = focus
@@ -233,76 +451,11 @@ class Example(object):
         """
         return any([self.context.focus, self.__dict__["focus"]])
 
-    @contextmanager
-    def execute_around_hooks(self, context_data):
-        around_generators = [
-            around(context_data) for around in self.context.all_around_functions
-        ]
-
-        for around_generator in reversed(around_generators):
-            next(around_generator)
-
-        yield
-
-        for around_generator in around_generators:
-            try:
-                next(around_generator)
-            except StopIteration:
-                pass
-
-    def _example_runner(self, context_data):
-        """
-        Execute before hooks, example and after hooks.
-        """
-        aggregated_exceptions = AggregatedExceptions()
-        with aggregated_exceptions.catch():
-            for before in self.context.all_before_functions:
-                before(context_data)
-            self.code(context_data)
-        after_functions = []
-        after_functions.extend(self.context.all_after_functions)
-        after_functions.extend(context_data.after_functions)
-        for after in reversed(after_functions):
-            with aggregated_exceptions.catch():
-                after(context_data)
-        for assertion in self.context.assertions:
-            with aggregated_exceptions.catch():
-                assertion()
-        if aggregated_exceptions.exceptions:
-            aggregated_exceptions.raise_correct_exception()
-
-    def _run_example(self, around_functions, context_data):
-        """
-        Run example, including all hooks.
-        """
-        if not around_functions:
-            self._example_runner(context_data)
-            return
-        around = around_functions.pop()
-
-        def wrapped():
-            self._run_example(around_functions, context_data)
-
-        around(context_data, wrapped)
-
     def __call__(self):
         """
         Run the example, including all around, before and after hooks.
         """
-        try:
-            if self.skip:
-                raise Skip()
-            _run_before_once_hooks()
-            context_data = _ContextData(self.context)
-            with _add_traceback_context_manager():
-                self._run_example(
-                    list(reversed(self.context.all_around_functions)), context_data
-                )
-        finally:
-            sys.stdout.flush()
-            sys.stderr.flush()
-            testslide.mock_callable.unpatch_all_callable_mocks()
-            testslide.mock_constructor.unpatch_all_constructor_mocks()
+        _ExampleRunner(self).run()
 
     def __str__(self):
         return self.name
@@ -320,9 +473,6 @@ class _TestSlideTestResult(unittest.TestResult):
 
     def _add_exception(self, err):
         exc_type, exc_value, exc_traceback = err
-        # Decorate __traceback__ for Python 2 compatibility
-        if sys.version_info[0] == 2:
-            exc_value.__dict__["__traceback__"] = err[2]
         self.aggregated_exceptions.append_exception(exc_value)
 
     def addError(self, test, err):
@@ -368,24 +518,7 @@ class Context(object):
     _SAME_CONTEXT_NAME_ERROR = "A context with the same name is already defined"
 
     # List of all top level contexts created
-    all_top_level_contexts = []  # type: List[Context]
-
-    def _setup_mock_callable(self):
-        def _mock_callable(self, *args, **kwargs):
-            return testslide.mock_callable.mock_callable(*args, **kwargs)
-
-        self.add_function("mock_callable", _mock_callable)
-
-        def register_assertion(assertion):
-            self.assertions.append(assertion)
-
-        testslide.mock_callable.register_assertion = register_assertion
-
-    def _setup_mock_constructor(self):
-        def _mock_constructor(self, *args, **kwargs):
-            return testslide.mock_constructor.mock_constructor(*args, **kwargs)
-
-        self.add_function("mock_constructor", _mock_constructor)
+    all_top_level_contexts = []
 
     # Constructor
 
@@ -412,7 +545,6 @@ class Context(object):
         self.examples = []
         self.before_functions = []
         self.after_functions = []
-        self.assertions = []
         self.around_functions = []
         self.context_data_methods = {}
         self.context_data_memoizable_attributes = {}
@@ -422,9 +554,6 @@ class Context(object):
         if not self.parent_context and not self.shared:
             self.all_top_level_contexts.append(self)
 
-        self._setup_mock_callable()
-        self._setup_mock_constructor()
-
     # Properties
 
     @property
@@ -595,11 +724,12 @@ class Context(object):
 
         return self.examples[-1]
 
-    def _context_data_has_attr(self, name):
+    def has_attribute(self, name):
         return any(
             [
                 name in self.context_data_methods.keys(),
                 name in self.context_data_memoizable_attributes.keys(),
+                name in self._runtime_attributes,
             ]
         )
 
@@ -607,7 +737,7 @@ class Context(object):
         """
         Add given function to example execution scope.
         """
-        if self._context_data_has_attr(name):
+        if self.has_attribute(name):
             raise AttributeError(
                 'Attribute "{}" already set for context "{}"'.format(name, self)
             )
@@ -629,7 +759,7 @@ class Context(object):
         Add given attribute name to execution scope, by lazily memoizing the return
         value of memoizable_code().
         """
-        if self._context_data_has_attr(name):
+        if self.has_attribute(name):
             raise AttributeError(
                 'Attribute "{}" already set for context "{}"'.format(name, self)
             )
@@ -659,16 +789,9 @@ class Context(object):
 
             # Build a child class of given TestCase, with a defined test that
             # will run TestSlide example.
-            if sys.version_info[0] == 2:
-                test_slide_test_case = type(
-                    str("TestSlideTestCase"),
-                    (test_case,),
-                    {"test_test_slide": test_test_slide},
-                )
-            else:
-                test_slide_test_case = types.new_class(
-                    "TestSlideTestCase", bases=(test_case,), exec_body=exec_body
-                )
+            test_slide_test_case = types.new_class(
+                "TestSlideTestCase", bases=(test_case,), exec_body=exec_body
+            )
 
             # This suite will only contain TestSlide's example test.
             test_suite = unittest.TestLoader().loadTestsFromName(
@@ -683,35 +806,11 @@ class Context(object):
         self.around_functions.append(wrap_test_case)
 
 
-before_once_functions = []  # type: List[function]
-before_once_executed = False
-
-
-def _run_before_once_hooks():
-    global before_once_executed
-    if not before_once_executed:
-        global before_once_functions
-        with _add_traceback_context_manager():
-            for code in before_once_functions:
-                code()
-        before_once_executed = True
-
-
 def reset():
     """
     Clear all defined contexts and hooks.
     """
-    if sys.version_info[0] >= 3:
-        Context.all_top_level_contexts.clear()
-    else:
-        Context.all_top_level_contexts[:] = []
-    global before_once_functions
-    if sys.version_info[0] >= 3:
-        before_once_functions.clear()
-    else:
-        before_once_functions[:] = []
-    global before_once_executed
-    before_once_executed = False
+    Context.all_top_level_contexts.clear()
 
 
 class TestCase(unittest.TestCase):
@@ -725,17 +824,23 @@ class TestCase(unittest.TestCase):
         )
         self.addCleanup(testslide.mock_callable.unpatch_all_callable_mocks)
         self.addCleanup(testslide.mock_constructor.unpatch_all_constructor_mocks)
+        self.addCleanup(testslide.patch_attribute.unpatch_all_mocked_attributes)
         super(TestCase, self).setUp()
 
     @staticmethod
     def mock_callable(target, method):
         return testslide.mock_callable.mock_callable(target, method)
 
+    @staticmethod
+    def mock_async_callable(target, method, callable_returns_coroutine=False):
+        return testslide.mock_callable.mock_async_callable(
+            target, method, callable_returns_coroutine
+        )
+
     @staticmethod
     def mock_constructor(target, class_name):
         return testslide.mock_constructor.mock_constructor(target, class_name)
 
-
-def _test_function(arg1, arg2, kwarg1=None, kwarg2=None):
-    "This function is used by some unit tests only"
-    return "original response"
+    @staticmethod
+    def patch_attribute(target, attribute, new_value):
+        return testslide.patch_attribute.patch_attribute(target, attribute, new_value)
