@@ -3,45 +3,33 @@
 # This source code is licensed under the MIT license found in the
 # LICENSE file in the root directory of this source tree.
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
-import six
+import asyncio
 import inspect
 import functools
 from typing import List, Callable  # noqa
 import testslide
 from testslide.strict_mock import StrictMock
 from testslide.strict_mock import _add_signature_validation
+from .patch import _patch, _is_instance_method
+from .lib import _bail_if_private
 
 
-def mock_callable(target, method):
-    if method == "__new__":
-        raise ValueError(
-            "Mocking __new__ is not allowed with mock_callable(), please use "
-            "mock_constructor()."
-        )
-    return _MockCallableDSL(target, method)
+def mock_callable(target, method, allow_private=False):
+    return _MockCallableDSL(target, method, allow_private=allow_private)
+
+
+def mock_async_callable(
+    target, method, callable_returns_coroutine=False, allow_private=False
+):
+    return _MockAsyncCallableDSL(
+        target, method, callable_returns_coroutine, allow_private
+    )
 
 
 _unpatchers = []  # type: List[Callable]  # noqa T484
 
 
-def unpatch_all_callable_mocks():
-    """
-    This method must be called after every test unconditionally to remove all
-    active mock_callable() patches.
-    """
-    try:
-        for unpatcher in _unpatchers:
-            unpatcher()
-    finally:
-        del _unpatchers[:]
-
-
-def register_assertion(assertion):
+def _default_register_assertion(assertion):
     """
     This method must be redefined by the test framework using mock_callable().
     It will be called when a new assertion is defined, passing a callable as an
@@ -51,6 +39,42 @@ def register_assertion(assertion):
     raise NotImplementedError("This method must be redefined by the test framework")
 
 
+register_assertion = _default_register_assertion
+_call_order_assertion_registered = False
+_received_ordered_calls = []
+_expected_ordered_calls = []
+
+
+def unpatch_all_callable_mocks():
+    """
+    This method must be called after every test unconditionally to remove all
+    active mock_callable() patches.
+    """
+    global register_assertion, _default_register_assertion, _call_order_assertion_registered, _received_ordered_calls, _expected_ordered_calls
+
+    register_assertion = _default_register_assertion
+    _call_order_assertion_registered = False
+    del _received_ordered_calls[:]
+    del _expected_ordered_calls[:]
+
+    unpatch_exceptions = []
+    for unpatcher in _unpatchers:
+        try:
+            unpatcher()
+        except Exception as e:
+            unpatch_exceptions.append(e)
+    del _unpatchers[:]
+    if unpatch_exceptions:
+        raise RuntimeError(
+            "Exceptions raised when unpatching: {}".format(unpatch_exceptions)
+        )
+
+
+def _is_setup():
+    global register_assertion, _default_register_assertion
+    return register_assertion is not _default_register_assertion
+
+
 def _format_target(target):
     if hasattr(target, "__repr__"):
         return repr(target)
@@ -62,17 +86,30 @@ def _format_target(target):
 
 def _format_args(indent, *args, **kwargs):
     indentation = "  " * indent
-    s = ("{}{}\n" "{}").format(indentation, args, indentation)
-    s += "{"
+    s = ""
+    if args:
+        s += ("{}{}\n").format(indentation, args)
     if kwargs:
-        s += "\n"
-        for k in sorted(kwargs.keys()):
-            s += "{}  {}={},\n".format(indentation, k, kwargs[k])
-        s += "{}".format(indentation)
-    s += "}\n"
+        s += indentation + "{"
+        if kwargs:
+            s += "\n"
+            for k in sorted(kwargs.keys()):
+                s += "{}  {}={},\n".format(indentation, k, kwargs[k])
+            s += "{}".format(indentation)
+        s += "}\n"
     return s
 
 
+def _is_coroutine(obj):
+
+    return inspect.iscoroutine(obj) or isinstance(obj, asyncio.coroutines.CoroWrapper)
+
+
+##
+## Exceptions
+##
+
+
 class UndefinedBehaviorForCall(BaseException):
     """
     Raised when a mock receives a call for which no behavior was defined.
@@ -97,28 +134,65 @@ class UnexpectedCallArguments(BaseException):
     """
 
 
+class NotACoroutine(BaseException):
+    """
+    Raised when a mock that requires a coroutine is not mocked with one.
+
+    Inherits from BaseException to avoid being caught by tested code.
+    """
+
+
 ##
-## Behavior
+## Runners
 ##
 
 
-class _Runner(object):
+class _BaseRunner:
     def __init__(self, target, method, original_callable):
         self.target = target
         self.method = method
         self.original_callable = original_callable
         self.accepted_args = None
-        self.call_count = 0
-        self.max_calls = None
 
-    def run(self, *args, **kwargs):
-        self.call_count += 1
-        if self.max_calls and self.call_count > self.max_calls:
+        self._call_count = 0
+        self._max_calls = None
+        self._has_order_assertion = False
+
+    def register_call(self, *args, **kwargs):
+        global _received_ordered_calls
+
+        if self._has_order_assertion:
+            _received_ordered_calls.append((self.target, self.method, self))
+
+        self.inc_call_count()
+
+    @property
+    def call_count(self):
+        return self._call_count
+
+    @property
+    def max_calls(self):
+        return self._max_calls
+
+    def _set_max_calls(self, times):
+        if not self._max_calls or times < self._max_calls:
+            self._max_calls = times
+
+    def inc_call_count(self):
+        self._call_count += 1
+        if self.max_calls and self._call_count > self.max_calls:
             raise UnexpectedCallReceived(
                 (
-                    "{}, {}: Unexpected call received.\n"
-                    "  Expected to receive at most {} calls, but an extra call was made."
-                ).format(_format_target(self.target), repr(self.method), self.max_calls)
+                    "Unexpected call received.\n"
+                    "{}, {}:\n"
+                    "  expected to receive at most {} calls with {}"
+                    "  but received an extra call."
+                ).format(
+                    _format_target(self.target),
+                    repr(self.method),
+                    self.max_calls,
+                    self._args_message(),
+                )
             )
 
     def add_accepted_args(self, *args, **kwargs):
@@ -141,10 +215,6 @@ class _Runner(object):
         else:
             return "any arguments "
 
-    def _set_max_calls(self, times):
-        if not self.max_calls or (self.max_calls and times < self.max_calls):
-            self.max_calls = times
-
     def add_exact_calls_assertion(self, times):
         self._set_max_calls(times)
 
@@ -209,14 +279,72 @@ class _Runner(object):
 
         register_assertion(assertion)
 
+    def add_call_order_assertion(self):
+        global _call_order_assertion_registered, _received_ordered_calls, _expected_ordered_calls
+
+        if not _call_order_assertion_registered:
+
+            def assertion():
+                if _received_ordered_calls != _expected_ordered_calls:
+                    raise AssertionError(
+                        (
+                            "calls did not match assertion.\n"
+                            "\n"
+                            "These calls were expected to have happened in order:\n"
+                            "\n"
+                            "{}\n"
+                            "\n"
+                            "but these calls were made:\n"
+                            "\n"
+                            "{}"
+                        ).format(
+                            "\n".join(
+                                (
+                                    "  {}, {} with {}".format(
+                                        _format_target(target),
+                                        repr(method),
+                                        runner._args_message().rstrip(),
+                                    )
+                                    for target, method, runner in _expected_ordered_calls
+                                )
+                            ),
+                            "\n".join(
+                                (
+                                    "  {}, {} with {}".format(
+                                        _format_target(target),
+                                        repr(method),
+                                        runner._args_message().rstrip(),
+                                    )
+                                    for target, method, runner in _received_ordered_calls
+                                )
+                            ),
+                        )
+                    )
+
+            register_assertion(assertion)
+            _call_order_assertion_registered = True
+
+        _expected_ordered_calls.append((self.target, self.method, self))
+        self._has_order_assertion = True
+
+
+class _Runner(_BaseRunner):
+    def run(self, *args, **kwargs):
+        super().register_call(*args, **kwargs)
+
+
+class _AsyncRunner(_BaseRunner):
+    async def run(self, *args, **kwargs):
+        super().register_call(*args, **kwargs)
+
 
 class _ReturnValueRunner(_Runner):
     def __init__(self, target, method, original_callable, value):
-        super(_ReturnValueRunner, self).__init__(target, method, original_callable)
+        super().__init__(target, method, original_callable)
         self.return_value = value
 
     def run(self, *args, **kwargs):
-        super(_ReturnValueRunner, self).run(*args, **kwargs)
+        super().run(*args, **kwargs)
         return self.return_value
 
 
@@ -251,9 +379,6 @@ class _YieldValuesRunner(_Runner):
         self.index += 1
         return item
 
-    if six.PY2:
-        next = __next__
-
     def run(self, *args, **kwargs):
         super(_YieldValuesRunner, self).run(*args, **kwargs)
         return self
@@ -279,22 +404,70 @@ class _ImplementationRunner(_Runner):
         return self.new_implementation(*args, **kwargs)
 
 
+class _AsyncImplementationRunner(_AsyncRunner):
+    def __init__(self, target, method, original_callable, new_implementation):
+        super().__init__(target, method, original_callable)
+        self.new_implementation = new_implementation
+
+    async def run(self, *args, **kwargs):
+        await super().run(*args, **kwargs)
+        coro = self.new_implementation(*args, **kwargs)
+        if not _is_coroutine(coro):
+            raise NotACoroutine(
+                f"Function did not return a coroutine.\n"
+                f"{self.new_implementation} must return a coroutine."
+            )
+        return await coro
+
+
 class _CallOriginalRunner(_Runner):
     def run(self, *args, **kwargs):
         super(_CallOriginalRunner, self).run(*args, **kwargs)
         return self.original_callable(*args, **kwargs)
 
 
+class _AsyncCallOriginalRunner(_AsyncRunner):
+    async def run(self, *args, **kwargs):
+        await super().run(*args, **kwargs)
+        return await self.original_callable(*args, **kwargs)
+
+
+##
+## Callable Mocks
+##
+
+
 class _CallableMock(object):
-    def __init__(self, target, method):
+    def __init__(self, target, method, is_async=False):
         self.target = target
         self.method = method
         self.runners = []
+        self.is_async = is_async
 
-    def __call__(self, *args, **kwargs):
+    def _get_runner(self, *args, **kwargs):
         for runner in self.runners:
             if runner.can_accept_args(*args, **kwargs):
-                return runner.run(*args, **kwargs)
+                return runner
+        return None
+
+    def __call__(self, *args, **kwargs):
+        runner = self._get_runner(*args, **kwargs)
+        if runner:
+            if isinstance(runner, _AsyncRunner):
+
+                async def await_runner(*args, **kwargs):
+                    return await runner.run(*args, **kwargs)
+
+                return await_runner(*args, **kwargs)
+            else:
+                if self.is_async:
+
+                    async def sync_wrapper(*args, **kwargs):
+                        return runner.run(*args, **kwargs)
+
+                    return sync_wrapper(*args, **kwargs)
+                else:
+                    return runner.run(*args, **kwargs)
         ex_msg = (
             "{}, {}:\n"
             "  Received call:\n"
@@ -320,136 +493,150 @@ class _CallableMock(object):
         return [runner.accepted_args for runner in self.runners if runner.accepted_args]
 
 
-class _DescriptorProxy(object):
-    def __init__(self, original_class_attr, attr_name):
-        self.original_class_attr = original_class_attr
-        self.attr_name = attr_name
-        self.instance_attr_map = {}
-
-    def __set__(self, instance, value):
-        self.instance_attr_map[id(instance)] = value
-
-    def __get__(self, instance, owner):
-        if instance is None:
-            return self
-        if id(instance) in self.instance_attr_map:
-            return self.instance_attr_map[id(instance)]
-        else:
-            if self.original_class_attr:
-                return self.original_class_attr.__get__(instance, owner)
-            else:
-                for parent in owner.mro()[1:]:
-                    method = parent.__dict__.get(self.attr_name, None)
-                    if type(method) is type(self):
-                        continue
-                    if method:
-                        return method.__get__(instance, owner)
-                return instance.__get__(instance, owner)
-
-    def __delete__(self, instance):
-        if instance in self.instance_attr_map:
-            del self.instance_attr_map[instance]
-
-
-def _is_instance_method(target, method):
-    if inspect.ismodule(target):
-        return False
-
-    if inspect.isclass(target):
-        klass = target
-    else:
-        klass = type(target)
-
-    for k in klass.mro():
-        if method in k.__dict__:
-            value = k.__dict__[method]
-            if isinstance(value, _DescriptorProxy):
-                value = value.original_class_attr
-            if inspect.isfunction(value):
-                return True
-    return False
-
-
-def _mock_instance_attribute(instance, attr, value):
-    """
-    Patch attribute at instance with given value. This works for any instance
-    attribute, even when the attribute is defined via the descriptor protocol using
-    __get__ at the class (eg with @property).
-
-    This allows mocking of the attribute only at the desired instance, as opposed to
-    using Python's unittest.mock.patch.object + PropertyMock, that requires patching
-    at the class level, thus affecting all instances (not only the one you want).
-    """
-    klass = type(instance)
-    class_restore_value = klass.__dict__.get(attr, None)
-    setattr(klass, attr, _DescriptorProxy(class_restore_value, attr))
-    setattr(instance, attr, value)
-
-    def unpatch_class():
-        if class_restore_value:
-            setattr(klass, attr, class_restore_value)
-        else:
-            delattr(klass, attr)
-
-    return unpatch_class
-
-
-def _patch(target, method, new_value):
-    if isinstance(target, six.string_types):
-        target = testslide._importer(target)
-
-    if isinstance(target, StrictMock):
-        original_callable = None
-    else:
-        original_callable = getattr(target, method)
-
-    new_value = _add_signature_validation(new_value, target, method)
-    restore_value = target.__dict__.get(method, None)
-
-    if inspect.isclass(target):
-        if _is_instance_method(target, method):
-            raise ValueError(
-                "Patching an instance method at the class is not supported: "
-                "bugs are easy to introduce, as patch is not scoped for an "
-                "instance, which can potentially even break class behavior; "
-                "assertions on calls are ambiguous (for every instance or one "
-                "global assertion?)."
-            )
-        new_value = staticmethod(new_value)
-
-    if _is_instance_method(target, method):
-        unpatcher = _mock_instance_attribute(target, method, new_value)
-    else:
-        setattr(target, method, new_value)
-
-        def unpatcher():
-            if restore_value:
-                setattr(target, method, restore_value)
-            else:
-                delattr(target, method)
-
-    return original_callable, unpatcher
+##
+## Support
+##
 
 
 class _MockCallableDSL(object):
 
     CALLABLE_MOCKS = {}  # NOQA T484
 
+    def _validate_patch(
+        self,
+        name="mock_callable",
+        other_name="mock_async_callable",
+        coroutine_function=False,
+        callable_returns_coroutine=False,
+    ):
+        if self._method == "__new__":
+            raise ValueError(
+                f"Mocking __new__ is not allowed with {name}(), please use "
+                "mock_constructor()."
+            )
+        _bail_if_private(self._method, self.allow_private)
+        if isinstance(self._target, StrictMock):
+            template_value = getattr(self._target._template, self._method, None)
+            if template_value and callable(template_value):
+                if not coroutine_function and inspect.iscoroutinefunction(
+                    template_value
+                ):
+                    raise ValueError(
+                        f"{name}() can not be used with coroutine functions.\n"
+                        f"The attribute '{self._method}' of the template class "
+                        f"of {self._target} is a coroutine function. You can "
+                        f"use {other_name}() instead."
+                    )
+                if (
+                    coroutine_function
+                    and (
+                        # FIXME We can not reliably introspect coroutine functions
+                        # for builtins: https://bugs.python.org/issue38225
+                        type(template_value) is not type(list.append)
+                        and not inspect.iscoroutinefunction(template_value)
+                    )
+                    and not callable_returns_coroutine
+                ):
+                    raise ValueError(
+                        f"{name}() can not be used with non coroutine "
+                        "functions.\n"
+                        f"The attribute '{self._method}' of the template class "
+                        f"of {self._target} is not a coroutine function. You "
+                        f"can use {other_name}() instead."
+                    )
+        else:
+            if inspect.isclass(self._target) and _is_instance_method(
+                self._target, self._method
+            ):
+                raise ValueError(
+                    "Patching an instance method at the class is not supported: "
+                    "bugs are easy to introduce, as patch is not scoped for an "
+                    "instance, which can potentially even break class behavior; "
+                    "assertions on calls are ambiguous (for every instance or one "
+                    "global assertion?)."
+                )
+            original_callable = getattr(self._target, self._method)
+            if not callable(original_callable):
+                raise ValueError(
+                    f"{name}() can only be used with callable attributes and "
+                    f"{repr(original_callable)} is not."
+                )
+            if inspect.isclass(original_callable):
+                raise ValueError(
+                    f"{name}() can not be used with with classes: "
+                    f"{repr(original_callable)}. Perhaps you want to use "
+                    "mock_constructor() instead."
+                )
+            if not coroutine_function and inspect.iscoroutinefunction(
+                original_callable
+            ):
+                raise ValueError(
+                    f"{name}() can not be used with coroutine functions.\n"
+                    f"{original_callable} is a coroutine function. You can use "
+                    f"{other_name}() instead."
+                )
+            if (
+                coroutine_function
+                and (
+                    # FIXME We can not reliably introspect coroutine functions
+                    # for builtins: https://bugs.python.org/issue38225
+                    type(original_callable) is not type(list.append)
+                    and not inspect.iscoroutinefunction(original_callable)
+                )
+                and not callable_returns_coroutine
+            ):
+                raise ValueError(
+                    f"{name}() can not be used with non coroutine functions.\n"
+                    f"{original_callable} is not a coroutine function. You can "
+                    f"use {other_name}() instead."
+                )
+
+    def _patch(self, new_value):
+        self._validate_patch()
+
+        if isinstance(self._target, StrictMock):
+            original_callable = None
+        else:
+            original_callable = getattr(self._target, self._method)
+
+        if not isinstance(self._target, StrictMock):
+            new_value = _add_signature_validation(new_value, self._target, self._method)
+
+        restore = self._method in self._target.__dict__
+        restore_value = self._target.__dict__.get(self._method, None)
+
+        if inspect.isclass(self._target):
+            new_value = staticmethod(new_value)
+
+        unpatcher = _patch(
+            self._target, self._method, new_value, restore, restore_value
+        )
+
+        return original_callable, unpatcher
+
+    def _get_callable_mock(self):
+        return _CallableMock(self._original_target, self._method)
+
     def __init__(
         self,
         target,
         method,
         callable_mock=None,
         original_callable=None,
-        prepend_first_arg=None,
+        allow_private=False,
     ):
+        if not _is_setup():
+            raise RuntimeError(
+                "mock_callable() was not setup correctly before usage. "
+                "Please refer to the documentation at http://testslide.readthedocs.io/ "
+                "to learn how to properly do that."
+            )
         self._original_target = target
         self._method = method
         self._runner = None
         self._next_runner_accepted_args = None
-        self.prepend_first_arg = prepend_first_arg
-
-        if isinstance(target, six.string_types):
+        self.allow_private = allow_private
+        if isinstance(target, str):
             self._target = testslide._importer(target)
         else:
             self._target = target
@@ -459,7 +646,7 @@ class _MockCallableDSL(object):
         if target_method_id not in self.CALLABLE_MOCKS:
             if not callable_mock:
                 patch = True
-                callable_mock = _CallableMock(self._original_target, self._method)
+                callable_mock = self._get_callable_mock()
             else:
                 patch = False
             self.CALLABLE_MOCKS[target_method_id] = callable_mock
@@ -471,7 +658,7 @@ class _MockCallableDSL(object):
             _unpatchers.append(del_callable_mock)
 
             if patch:
-                original_callable, unpatcher = _patch(target, method, callable_mock)
+                original_callable, unpatcher = self._patch(callable_mock)
                 _unpatchers.append(unpatcher)
             self._original_callable = original_callable
             callable_mock.original_callable = original_callable
@@ -510,8 +697,6 @@ class _MockCallableDSL(object):
         """
         Filter for only calls like this.
         """
-        if self.prepend_first_arg:
-            args = (self.prepend_first_arg,) + args
         if self._runner:
             self._runner.add_accepted_args(*args, **kwargs)
         else:
@@ -619,11 +804,6 @@ class _MockCallableDSL(object):
 
         @functools.wraps(func)
         def wrapper(*args, **kwargs):
-            if self.prepend_first_arg and args:
-                assert (
-                    args[0] == self.prepend_first_arg
-                ), "Received unexpected first argument: {}.".format(args[0])
-                args = args[1:]
             return func(self._original_callable, *args, **kwargs)
 
         self._add_runner(
@@ -659,19 +839,17 @@ class _MockCallableDSL(object):
         If assertion is for 0 calls, any received call will raise
         UnexpectedCallReceived and also an AssertionError.
         """
-        if count is 0:
-            if self._runner:
-                raise ValueError(
-                    "Asked to not accept any calls, but a behavior was previously defined."
-                )
-            self.to_raise(
-                UnexpectedCallReceived(
-                    ("{}, {}: Excepted not to be called!").format(
-                        _format_target(self._target), repr(self._method)
+        if count:
+            self._assert_runner()
+        else:
+            if not self._runner:
+                self.to_raise(
+                    UnexpectedCallReceived(
+                        ("{}, {}: Expected not to be called!").format(
+                            _format_target(self._target), repr(self._method)
+                        )
                     )
                 )
-            )
-        self._assert_runner()
         self._runner.add_exact_calls_assertion(count)
         return self
 
@@ -718,3 +896,89 @@ class _MockCallableDSL(object):
         Short for self.and_assert_called_at_least(1).
         """
         return self.and_assert_called_at_least(1)
+
+    def and_assert_called_ordered(self):
+        """
+        Assert that multiple calls, potentially to different mock_callable()
+        targets, happened in the order defined.
+        """
+        self._assert_runner()
+        self._runner.add_call_order_assertion()
+        return self
+
+
+class _MockAsyncCallableDSL(_MockCallableDSL):
+    def __init__(self, target, method, callable_returns_coroutine, allow_private=False):
+        self._callable_returns_coroutine = callable_returns_coroutine
+        super().__init__(target, method, allow_private=allow_private)
+
+    def _validate_patch(self):
+        return super()._validate_patch(
+            name="mock_async_callable",
+            other_name="mock_callable",
+            coroutine_function=True,
+            callable_returns_coroutine=self._callable_returns_coroutine,
+        )
+
+    def _get_callable_mock(self):
+        return _CallableMock(self._original_target, self._method, is_async=True)
+
+    def with_implementation(self, func):
+        """
+        Replace callable by given async function.
+        """
+        if not callable(func):
+            raise ValueError("{} must be callable.".format(func))
+        self._add_runner(
+            _AsyncImplementationRunner(
+                self._original_target, self._method, self._original_callable, func
+            )
+        )
+        return self
+
+    def with_wrapper(self, func):
+        """
+        Replace callable with given wrapper async function, that will be called as:
+
+          await func(original_async_func, *args, **kwargs)
+
+        receiving the original function as the first argument as well as any given
+        arguments.
+        """
+        if not callable(func):
+            raise ValueError("{} must be callable.".format(func))
+
+        if not self._original_callable:
+            raise ValueError("Can not wrap original callable that does not exist.")
+
+        @functools.wraps(func)
+        async def wrapper(*args, **kwargs):
+            coro = func(self._original_callable, *args, **kwargs)
+            if not _is_coroutine(coro):
+                raise NotACoroutine(
+                    f"Function did not return a coroutine.\n"
+                    f"{func} must return a coroutine."
+                )
+            return await coro
+
+        self._add_runner(
+            _AsyncImplementationRunner(
+                self._original_target, self._method, self._original_callable, wrapper
+            )
+        )
+        return self
+
+    def to_call_original(self):
+        """
+        Calls the original callable implementation, instead of mocking it. This is
+        useful for example, if you want to by default call the original implementation,
+        but for a specific calls, mock the result.
+        """
+        if not self._original_callable:
+            raise ValueError("Can not call original callable that does not exist.")
+        self._add_runner(
+            _AsyncCallOriginalRunner(
+                self._original_target, self._method, self._original_callable
+            )
+        )
+        return self
